/** API to create exercises */

import { getTerminalContents, getRawTerminalContents, runCommand, runUnexecutedCommand, abort, isVmLoading } from "./jslinux.api.mjs"
import { arrayEquals } from "./util.mjs"
import { getHandInToken } from "./snakeoil.mjs"
import { State } from "./state.mjs"

/**
 * An exercise. It has a human-readable name and an `execute` function
 * that contains the logic of the exercise.
 * 
 * The execute function is run bound to an ExerciseExecutionContext, which
 * provides API methods the teacher can call when creating exercises in
 * exercises.mjs.
 * 
 * Instances of this class are not created directly by the user, instead,
 * they are created using the `exercise` function.
 */
export class Exercise {
    #execute = null
    #context = new ExerciseExecutionContext()

    /**
     * @param {string} name Human-readable name of the exercise
     * @param {number} index Index of the exercise in the list of exercises of the test
     * @param {Function} execute A function to configure and execute this exercise
     */
    constructor(name, index, execute) {
        this.name = name
        this.index = index
        this.#execute = execute
    }
    
    /**
     * Start the exercise.
     */
    start() {
        // Bind the execute function (that the teacher writes to create an exercise,
        // see exercises.mjs) to the ExerciseExecutionContext.
        // This means that in the execute function, the `this` keyword will point to
        // said context.
        this.#execute.bind(this.#context)()
    }

    /**
     * Allow the user to retry the exercise after they failed, reusing their
     * current terminal session etc.
     * 
     * This is accomplished by restarting the exercise in a context that
     * disables e.g. prepareWith, but normally allows awaiting manual confirmation.
     */
    retry() {
        this.#context = new ExerciseRetryExecutionContext(this.#context)
        this.start()
    }

    /**
     * Inform the execute function that the user has manually confirmed that
     * they are done with the exercise.
     */
    manuallyConfirm() {
        this.#context._confirmExerciseCompletion()
    }

    /**
     * Set the handler that is called every time the exercise executor
     * sets the description of the exercise.
     * @param {Function} handler Function that takes a description as parameter
     */
    onDescribed(handler) {
        this.#context._describeHandler = handler
    }

    /**
     * Set the handler that is called every time the exercise's completion
     * is in the process of being verified.
     * @param {Function} handler Function with no parameter
     */
    onVerifying(handler) {
        this.#context._verifyingHandler = handler
    }

    /**
     * Set the handler that is called every time the exercise's completion
     * is verified (this verification either failed or was successful)
     * @param {Function} handler Function that takes the result of the verification (true/false) as parameter 
     */
    onVerified(handler) {
        this.#context._verifyHandler = handler
    }

    /**
     * Get if the exercise is solved in the given current state
     * @param {State} state The current state
     */
    isSolved(state) {
        return state.solvedExercises.includes(this.index)
    }

    /**
     * Get if the exercise is current in the given current state
     * @param {State} state The current state
     */
    isCurrent(state) {
        return state.exerciseIndex === this.index
    }

    /**
     * Marks this exercise as solved, modifying the given current state
     * @param {State} state The current state
     */
    markAsSolved(state) {
        state.markAsSolved(this.index)
    }

    /**
     * Marks this exercise as current, modifying the given current state
     * @param {State} state The current state
     */
    markAsCurrent(state) {
        state.exerciseIndex = this.index
    }
}

/**
 * The context in which the exercise execution function is run.
 * 
 * It provides API methods the teacher can use while creating the
 * exercises in exercises.mjs. Documentation in this file is written
 * from the teacher's perspective.
 */
class ExerciseExecutionContext {
    constructor() {
        this._describeHandler = null
        this._verifyHandler = null
        this._verifyingHandler = null
        this._confirmExerciseCompletion = null
    }

    /**
     * Set the description of the exercise. After reading this, the student
     * should know exactly what is expected of them and can start thinking
     * about a solution.
     * 
     * @param {string} description Description of the exercise
     */
    describe(description) {
        this._describeHandler(description)
    }

    /**
     * Prepare the exercise by running the provided command before the
     * user is allowed to input anything.
     * 
     * By default, after running the provided command, "clear" is
     * executed to clear the terminal. If you do not want to clear the
     * terminal, set `keepVisible` to true.
     * 
     * If no command is provided (i.e. command === null), no command is executed.
     * This is helpful if you just want to clear the terminal at the start of an exercise.
     * @param {string} command
     * @param {boolean} keepVisible
     */
    prepareWith(command, keepVisible) {
        return this.#prepareWith(command, keepVisible)
    }

    /**
     * Private method actually doing all the work of prepareWith.
     * 
     * Needed because the prepareWith public API is disabled in retry contexts, but we still need
     * it internally.
     * @param {string} command
     * @param {boolean} keepVisible
     */
    #prepareWith(command, keepVisible) {
        return new Promise((resolve, _) => {
            // First, we define two helper functions:
            // 1. prepareAndResolve, which runs the preparation command, optionally clears the terminal and resolves the promise
            // 2. tryPrepareAndResolve, which tries to ensure that we can actually run commands, and then calls prepareAndResolve 
            async function prepareAndResolve() {
                // Actually prepare and resolve the promise (we get the resolve() function via the closure)
                if(command !== null) {
                    await runCommand(command)
                }
                if (!keepVisible) {
                    await runCommand("clear")
                }
                resolve()
            }

            let intervalId
            function tryPrepareAndResolve() {
                if(getTerminalContents().length === 0 && !isVmLoading()) {
                    // We don't have a prompt, and the VM doesn't appear to be loading.
                    // So maybe the user has opened a pager. Let's try to get out of this, using Ctrl+C.
                    abort()
                }

                if(getTerminalContents().length === 0 && !isVmLoading()) {
                    // If there is still no prompt, and just like before, the VM isn't loading,
                    // Ctrl+C apparently didn't work to exit whatever we were in.

                    // So we are probably in Vi.

                    // Try exiting that.
                    runCommand(":q!")
                }

                if (getTerminalContents().length > 0) {
                    // We have a prompt, so we can prepare and resolve.
                    clearInterval(intervalId)
                    prepareAndResolve()
                    return
                }
            }
            
            // We will repeatedly attempt to prepare and resolve in 0.5s intervals, starting 0.5s from now...
            intervalId = setInterval(tryPrepareAndResolve, 500)
            // ... and once now.
            tryPrepareAndResolve()
        })
    }

    /**
     * Clear the terminal. Shorthand for prepareWith(null, false)
     */
    async clear() {
        await this.#clear()
    }

    /**
     * Private method actually clearing the terminal. Again, this is needed because the clear() public API
     * method is disabled in retry contexts, but we still need it internally.
     */
    async #clear() {
        await this.#prepareWith(null, false)
    }

    /**
     * Indicates that manual confirmation is needed. Used like this:
     * 
     * ```javascript
     * await this.manualConfirmation()
     * ```
     * 
     * The code in the exercise execution function will continue only when
     * the user indicated that they think they are done with the exercise.
     * 
     * @returns Promise that is resolved when the user manually confirms they are done
     */
    manualConfirmation() {
        return new Promise((resolve, _) => {
            this._confirmExerciseCompletion = async () => {
                this._verifyingHandler()
                await runUnexecutedCommand()
                resolve()
            }
        })
    }

    /**
     * Verify the completion of the exercise by optionally executing a command and checking
     * if the output is as expected. If yes, the exercise is treated as solved, if no,
     * the user has to try again.
     * 
     * Used like this:
     * 
     * ```javascript
     * this.verify("cat botschaft").hasOutput("Der Junge sieht den Mann mit dem Fernglas.")
     * ```
     * 
     * verify() can also be called without a command, in which case no command is executed – instead
     * the last command entered by the user (if any) is used.
     * 
     * When the parameter passed to `hasOutput` is a string, the verification succeeds iff
     * this string is the first and only line of output of the command.
     * When the parameter passed to `hasOutput` is a list, the verification succeeds iff
     * the output lines are exactly the ones in the list.
     * 
     * @param {string} command A command to execute
     * @returns An object that has the method "hasOutput"
     */
    verify(command) {
        // If a command is provided, store the returned promise (which is resolved as soon as the command was run),
        // if no command is provided, directly start with a resolved promise.
        const commandPromise = command ? runCommand(command) : Promise.resolve()
        const verifyHandler = this._verifyHandler
        const clear = this.#clear.bind(this)

        return {
            /**
             * Verify the output of the command matches a certain condition.
             * 
             * - If param is a string, check that the only output line of the command is equal to the string
             * - If param is a list, check that the output lines of the command are exactly equal to the list
             * - If param is a function, the function is called with the actual output lines as parameter, and we check it returns
             *   a truthy value.
             * 
             * @param {Function|String|Object} param String or list of strings to exactly match, or predicate function
             */
            hasOutput(param) {
                // Wait until the VM has processed the input and written some output to the terminal
                commandPromise.then(async () => {
                    const terminalContents = getTerminalContents()

                    if (command && terminalContents.at(-1).input !== command) {
                        // If a command was given to be executed, verify that it was executed
                        throw new Error(`Could not find expected verification command '${command}' in terminal.`)
                    }

                    // We start by assuming the result was a failure...
                    let result = false

                    // ... then, check every entry in the terminal (starting from the last) to see if we can be convinced otherwise
                    for(const entry of terminalContents.toReversed()) {

                        // Poor man's switch-case: depending on the "typeof param", a different verification predicate is chosen
                        const predicate = {
                            "function": param,
                            "string": (actualOutput) => arrayEquals(actualOutput, [param]),
                            "object": (actualOutput) => arrayEquals(actualOutput, param)
                        }[typeof param]
    
                        // !! converts any value to a boolean (not a special operator, just two ! (boolean not) operators in series)
                        result = !!predicate(entry.output || [])

                        if(result) {
                            break
                        }
                    }

                    // After that is done, return the result
                    await clear()
                    verifyHandler(result)
                })
            },

            /**
             * Verify that the terminal contains the given line somewhere.
             * "Somewhere" might be a user input, a prompt, a line in an editor or something.
             * 
             * Normally, you'll probably want to use hasOutput(...) instead, because it ensures that the expected
             * line is found in an actual command's output. This command here is likely only useful when you
             * want to accept solutions where the line is displayed by an editor or pager, at the cost of
             * the user being able to trick you by just typing the solution as input.
             *  
             * @param {String} line A line of text that should be contained in the terminal
             */
            terminalContains(line) {
                // Wait until the VM has processed the input and written some output to the terminal
                commandPromise.then(async () => {
                    const terminalContents = getRawTerminalContents()
                    const result = terminalContents.includes(line)
                    await clear()
                    verifyHandler(result)
                })
            },

            /**
             * Verify the command matches a certain condition.
             * 
             * - If param is a string, check that the command is exactly equal to param
             * - If param is a function, the function is called with the actual command as parameter, and we check it returns
             *   a truthy value.
             * 
             * @param {Function|String} param String to exactly match, or predicate function
             */
            commandWas(param) {
                commandPromise.then(async () => {
                    const terminalContents = getTerminalContents()
                    const latestEntry = terminalContents[terminalContents.length - 1]

                    // Poor man's switch-case: depending on the "typeof param", a different verification predicate is chosen
                    const predicate = {
                        "function": param,
                        "string": (actualInput) => actualInput === param,
                    }[typeof param]

                    const result = !!predicate(latestEntry.input)

                    await clear()
                    verifyHandler(result)
                })
            }
        }
    }
}

/**
 * Exercise execution contexts for retry attempts. Disables the following public API functions:
 * 
 * - describe()
 * - prepareWith()
 * - clear()
 * 
 * This is done because, when retrying, the exercise setup code is re-run completely, however,
 * we do not want to e.g. prepare the exercise again, as all the respective files already exist
 * from the previous try.
 */
class ExerciseRetryExecutionContext extends ExerciseExecutionContext {

    /**
     * Create a ExerciseRetryExecutionContext
     * @param {ExerciseExecutionContext} oldContext An old context from which handlers are inherited
     */
    constructor(oldContext) {
        super()
        this._describeHandler = oldContext._describeHandler
        this._verifyHandler = oldContext._verifyHandler
        this._verifyingHandler = oldContext._verifyingHandler
    }

    describe() {}

    async prepareWith() {}

    async clear() {}
}

/**
 * A test. It has a human-readable name and a list of exercises.
 */
class Test {

    /**
     * @param {string} name Human-readable name of the test
     * @param {Exercise[]} exercises The exercises this test consists of
     * @param {boolean} handInEnabled If handing in is enabled (usually yes, but maybe not for practice tests)
     * @param {string} welcomeHtml HTML code to show in the welcome dialog
     */
    constructor(name, exercises, handInEnabled, welcomeHtml) {
        this.name = name
        this.exercises = exercises
        this.handInEnabled = handInEnabled
        this.welcomeHtml = welcomeHtml
    }

    /**
     * Return the next unsolved exercise, given the current state that's given as parameter
     * @param {State} state The current state
     */
    getNextUnsolvedExercise(state) {
        if (state.solvedExercises.length === this.exercises.length) {
            return null
        } else {
            let nextUnsolvedExerciseIndex = state.exerciseIndex || 0
            while (state.solvedExercises.includes(nextUnsolvedExerciseIndex)) {
                nextUnsolvedExerciseIndex = (nextUnsolvedExerciseIndex + 1) % this.exercises.length
            }
    
            return this.exercises[nextUnsolvedExerciseIndex]
        }
    }

    /**
     * Return if the welcome message was already read in the current state
     * @param {State} state The current state
     */
    wasWelcomeRead(state) {
        // TODO Properly implement this for each test separately
        return !state.firstStart
    }

    /**
     * Marks the welcome message as read, modifying the given current state
     * @param {State} state The current state
     */
    markWelcomeAsRead(state) {
        // TODO Properly implement this for each test separately
        state.firstStart = false
    }

    /**
     * Get the hand-in token to hand in this test
     * @param {State} state The current state
     */
    getHandInToken(state) {
        // TODO Properly implement this for each test separately
        return getHandInToken(state)
    }
}

/**
 * Create a test. Returns the test object
 * 
 * @param {string} name A human-readable name of the test
 * @param {Function} exerciseCreationFunction A function that calls `exercise` to create exercise.
 * Note: This must be a function declared with the function keyword, arrow functions (()=>{...}) do not work.
 */
export function test(name, exerciseCreationFunction) {

    const testConfigCollector = {
        exercises: [],
        handInEnabled: true,
        welcomeHtml: null,

        /**
         * Add an exercise.
         * 
         * @param {string} name A short name of the exercise
         * @param {Function} execute A function that executes the exercise. 
         * Note: This must be a function declared with the function keyword, arrow functions (()=>{...}) do not work.
         */
        exercise(name, execute) {
            const exerciseObj = new Exercise(name, this.exercises.length, execute)
            this.exercises.push(exerciseObj)
        },

        /**
         * Instruct the UI to not show the hand-in button for this test, ever
         */
        disableHandIn() {
            this.handInEnabled = false
        },

        /**
         * Set the message shown in the welcome dialog
         * @param {string} html The HTML source code that forms the welcome message
         */
        welcome(html) {
            this.welcomeHtml = html
        }
    }

    exerciseCreationFunction.bind(testConfigCollector)()

    return new Test(name, testConfigCollector.exercises, testConfigCollector.handInEnabled, testConfigCollector.welcomeHtml)
}
