/** API to create cards for exercises */

import { createCard, setCardHoverTitle } from "./cards.mjs"
import { Exercise } from "./exercises.api.mjs"
import { createElementWithClass } from "./util.mjs"

/**
 * Changes the appearance of the given exercise card so it looks like it is solved
 * (this method does not care if the exercise is actually solved)
 * @param {HTMLElement} cardEl The .card element
 */
export function displayAsSolved(cardEl) {
    cardEl.classList.add("solved")
    setCardHoverTitle(cardEl, "Diese Aufgabe ist bereits gelöst, du kannst nicht mehr zu ihr springen.", false)
}

/**
 * Changes the appearance of the given exercise card so it looks like it is current
 * @param {HTMLElement} cardEl The .card element
 */
export function displayAsCurrent(cardEl) {
    cardEl.classList.remove("collapsed")
    setCardHoverTitle(cardEl, "Du kannst nicht zur aktuellen Aufgabe springen.", false)
}

/**
 * Changes the appearance of the given exercise card so it looks like it is not current
 * @param {HTMLElement} cardEl The .card element
 */
export function displayAsNonCurrent(cardEl) {
    cardEl.classList.add("collapsed")

    if (!cardEl.classList.contains("solved")) {
        setCardHoverTitle(cardEl, "Zu dieser Aufgabe springen", true)
    }
}

/**
 * Changes the appearance of the given exercise card so it looks like it is currently being verified
 * @param {HTMLElement} cardEl The .card element
 */
export function displayAsVerifying(cardEl) {
    cardEl.classList.remove("unusual")
    cardEl.classList.add("verifying")
}

/**
 * Changes the appearance of the given exercise card so it shows a message that the verification takes
 * an unusually long time. Only works when already displayed as verifying.
 * @param {HTMLElement} cardEl The .card element
 */
export function displayAsUnusuallyLongVerifying(cardEl) {
    cardEl.classList.add("unusual")
}

/**
 * Changes the appearance of the given exercise card so it looks normal (no verification indicator)
 * @param {HTMLElement} cardEl The .card element
 */
export function displayAsNonVerifying(cardEl) {
    cardEl.classList.remove("verifying", "unusual")
}

/**
 * Create a confirm and a reset button that can be appended to the actions section of
 * an exercise card.
 * @param {Function} manuallyConfirm Function to call to manually confirm this exercise
 * @returns List of buttons
 */
function createExerciseCardActionButtons(manuallyConfirm) {
    const confirmButton = createElementWithClass("button", "primary")
    confirmButton.textContent = "Ich denke, ich bin fertig."
    confirmButton.onclick = () => manuallyConfirm()

    const resetButton = document.createElement("button")
    resetButton.textContent = "Aufgabe neu beginnen"
    resetButton.onclick = () => location.reload()

    return [confirmButton, resetButton]
}

/**
 * Create a DOM card for the given exercise
 * @param {Exercise} exercise The exercise
 * @param {Function} onTitleClick Event handler for click on the title
 */
export function createExerciseCard(exercise, onTitleClick) {
    const cardEl = createCard(exercise.name, "#" + exercise.index)

    cardEl.classList.add("exercise")
    cardEl.dataset.exerciseIndex = exercise.index
    // Add Bootstrap Icon's 'bi' class to the heading, because it will apply styles to
    // the h2::before element, whose content we set automatically in CSS, depending on the
    // solved state of the exercise
    cardEl.querySelector(".title h2").classList.add("bi")

    const descriptionEl = createElementWithClass("p", "description")
    const actionsEl = createElementWithClass("div", "actions")
    actionsEl.append(...createExerciseCardActionButtons(
        exercise.manuallyConfirm.bind(exercise)
    ))
    const loaderEl = createElementWithClass("div", "loader")
    loaderEl.innerHTML = `
    <div class="loader-bg"></div>
    <div class="loader-animation-wrapper">
        <span class="loader-animation"></span>
    </div>
    <span class="loader-desc standard">Deine Lösung wird überprüft...</span>
    <span class="loader-desc unusual">Das braucht ungewöhnlich lange. <a href="#">Aufgabe neu beginnen?</a></span>
    `
    loaderEl.querySelector(".loader-desc.unusual a").onclick = () => location.reload()
    cardEl.querySelector(".content").append(descriptionEl, actionsEl, loaderEl)

    // When the title bar of an exercise is clicked, skip to that exercise
    const titleEl = cardEl.querySelector(".title")
    titleEl.addEventListener("click", onTitleClick)

    return cardEl
}
