/** API to interact with lib/jslinux **/

import { waitFor } from "./util.mjs"

/**
 * Try to parse the given line as if it contained a prompt, maybe followed by user input.
 * 
 * @param {string} line A line that might contain a prompt, followed by optional user input
 * @returns An object containing prompt and input, or null if the line wasn't a prompt line
 */
function parseAsPrompt(line) {
    const match = line.match(/(\/.* #)(?: (.*))?/)

    if (match !== null) {
        return {
            prompt: match[1],
            input: match[2]
        }
    } else {
        return null
    }
}

/**
 * Parse the given line as if it was a line of output.
 * 
 * @param {string} line A line containing output
 * @returns An object with an output property, which is a list containing only this line
 */
function parseAsOutput(line) {
    return {
        output: [line]
    }
}

/**
 * Try to parse the given line as if it contained a prompt, maybe followed by user input.
 * If this isn't successful, parse it as if it contained output.
 * 
 * @param {string} line A line containing either a prompt and optionally input or some output
 * @returns An object containing either prompt and input, or the output
 */
function parseAsPromptOrOutput(line) {
    const parsedAsPrompt = parseAsPrompt(line)

    if (parsedAsPrompt !== null) {
        return parsedAsPrompt
    } else {
        return parseAsOutput(line)
    }
}

/**
 * Merges two parses of a terminal line. Return an object with the following properties
 * created in the following way:
 * 
 * - prompt: parse1.prompt || parse2.prompt || null
 * - input: parse1.input || parse2.input || null
 * - output: If !parse1.output, parse1.input is treated as [].
 *   If !parse2.output, parse2.output is treated as [].
 *   The output property is then set to parse1.output.concat(parse2.output)
 * 
 * @param {Object} parse1 A parse of a terminal line
 * @param {Object} parse2 Another parse of a terminal line
 * @returns The two parses, merged
 */
function mergeParses(parse1, parse2) {
    return {
        prompt: parse1.prompt || parse2.prompt || null,
        input: parse1.input || parse2.input || null,
        output: (parse1.output || []).concat(parse2.output || [])
    }
}

/**
 * Sanitize a line read from a terminal.
 * 
 * Needed because the terminal uses non-breaking spaces everywhere (to avoid that
 * the browser "intelligently" collapses multiple spaces into one), but JavaScript
 * users just expect normal spaces (as " " !== "   " is true anyways).
 * @param {string} line
 */
function sanitize(line) {
    // Replace non-breaking spaces with normal ones
    return line.replaceAll("\u00a0", " ")
}

/**
 * Get the contents of the terminal area that are currently visible (without scrolling).
 * 
 * Empty lines are ignored. All other lines are scanned for prompts. If a line contains 
 * a prompt, it is treated as an input line, otherwise, as an output line.
 *
 * If an output line directly follows an input line, it corresponds to that input line.
 * If an output line directly follows an output line, it corresponds to the input the preceding
 * output line corresponds to.
 * It is possible for an output line to not correspond to any input line, which is the case
 * when either the input line is not visible anymore due to scrolling or there never has been
 * an input line (this happens after boot or when the user opens e.g. an editor).
 * 
 * The function returns a list of objects. Each object represents one input line and all
 * corresponding output lines.
 * 
 * ## Potential error sources
 * 
 * - False positives in prompt detection: If a command outputs something that looks like a prompt,
 *   or if a user opens an editor and types in a prompt, this line is wrongly detected as containing
 *   a prompt.
 * - Scrolling problems: If the output of a command is longer than the terminal (and/or suppresses the
 *   input line), the returned results won't contain all the output or the input line
 * - Empty lines are ignored: Even if a command legitimately outputs an empty line, it will be ignored.
 * 
 * @param includeLastInteraction If the last interaction (which only consists of a prompt, with nothing executed yet) should be included
 * @returns List of objects of the form {"prompt": "...", "input": "...", "output": ["...", "..."]}
 */
export function getTerminalContents(includeLastInteraction) {
    const contents = []
    let currentInteraction = {}

    getRawTerminalContents()
        .filter(line => line !== " ")
        .map(parseAsPromptOrOutput)
        .forEach(parsed => {
            if (parsed.hasOwnProperty("input")) {
                contents.push(currentInteraction)
                currentInteraction = parsed
            } else {
                currentInteraction = mergeParses(currentInteraction, parsed)
            }
        })

    if(includeLastInteraction) {
        contents.push(currentInteraction)
    }

    return contents
}

/**
 * Get the raw contents of the terminal area that are currently visible (without scrolling).
 * 
 * Returns a list containing every line (even empty ones) of the visible terminal area.
 * Just like getTerminalContents(), the lines are sanitized so that non-breaking spaces are
 * replaced with normal spaces.
 * 
 * @returns List of strings ["firstVisibleLine", "secondVisibleLine", ..., "lastVisibleLine"]
 */
export function getRawTerminalContents() {
    return Array
        .from(document.querySelectorAll(".term_content div"))
        .map(div => sanitize(div.textContent))
}

/**
 * Return if the vm is currently loading, i.e. not accepting commands
 */
export function isVmLoading() {
    const contents = getTerminalContents(true)
    return contents.length === 1 && contents[0].prompt === null && contents[0].input === null && contents[0].output[0] == "Loading..."
}

/**
 * Return the current prompt, or null, if no prompt is currently shown
 */
function getCurrentPrompt() {
    const lastInteraction = getTerminalContents(true).at(-1)
    if(lastInteraction.prompt && !lastInteraction.input?.trim()) {
        return lastInteraction.prompt
    } else {
        return null
    }
}

/**
 * Returns the current prompt as soon as one is shown
 */
function prompt() {
    return waitFor(getCurrentPrompt)
}

/**
 * Calls the given handler function without parameters whenever the user presses the key enter
 * @param {Function} handler A handler function
 */
export function onEnterPressed(handler) {
    document.querySelector(".term_textarea").addEventListener("keydown", (ev) => {
        if (ev.key === "Enter") {
            handler()
        }
    })
}

/**
 * Simulates a press of the enter button
 */
async function pressEnter() {
    const textarea = document.querySelector(".term_textarea")
    // We append \n to simulate the Enter press
    textarea.value += "\n"

    // Now we just have to trigger an input event to activate the JSLinux event handler, see
    // https://thewebdev.info/2021/05/02/how-to-programmatically-trigger-a-change-event-on-an-input-with-javascript/
    textarea.dispatchEvent(new Event("input"))

    // And wait until a prompt is shown (-> until the command is executed)
    await prompt()
}

/**
 * Simulates pressing Ctrl+C to send a SIGINT command to the currently running process
 * 
 * @param {boolean} waitForPrompt If true, wait until a prompt is shown, if false, exit immediately
 */
export async function abort(waitForPrompt = false) {
    const textarea = document.querySelector(".term_textarea")
    textarea.dispatchEvent(new KeyboardEvent("keydown", {
        keyCode: 67,
        ctrlKey: true
    }))

    if(waitForPrompt) {
        await prompt()
    }
}

/**
 * Types the given command into the terminal and simulates an Enter press
 * The command will then be executed, and the output will be shown in the terminal
 * @param {string} command A command to execute
 */
export async function runCommand(command) {
    document.querySelector(".term_textarea").value = command
    await pressEnter()
}

/**
 * Checks if the user has entered a command but not yet executed it. If that is the case,
 * press enter, and wait for it to complete.
 */
export async function runUnexecutedCommand() {
    const lastInteraction = getTerminalContents(true).at(-1)
    if(lastInteraction.prompt && lastInteraction.input?.trim()) {
        await pressEnter()
    }
}

/**
 * Enables (or disables) auto-focus to the terminal whenever a key is pressed.
 * This allows the user to click on an exercise completion button and then seamlessly continue
 * typing into the terminal, without having to click it first.
 * 
 * Note: This also prevents any shortcuts from working, so use with care
 */
export function setAutoFocusEnabled(enabled) {
    window.onkeydown = enabled ? function () {
        const textareaElement = document.querySelector("textarea.term_textarea")
        if (document.activeElement !== textareaElement) {
            textareaElement.focus()
        }
    } : null
}

/**
 * Register a function that is called every time the user wants to execute Ctrl+C or Ctrl+V in the terminal.
 * Does not prevent this from being executed!
 * 
 * @param {Function} handler Function to be called without parameters on Ctrl+C and Ctrl+V
 */
export function onCopyOrPasteHotkey(handler) {
    document.querySelector("textarea.term_textarea").addEventListener("keydown", event=>{
        console.log(event)
        if(event.ctrlKey && (event.key == "c" || event.key == "v")) {
            handler()
        }
    }, true)
}
