/** Read and modify the current state of the application. Allows saving to local storage. */

import { generateSessionUuid } from "./snakeoil.mjs"

export class State {

    /**
     * Create a new state object, backed by local storage. The realm parameter allows to store
     * several different states side-by-side, without interfering.
     * @param {string} realm
     */
    constructor(realm) {
        this.realm = realm
    }

    /**
     * Return a uuid for this test session
     */
    get sessionUuid() {
        let uuid = localStorage.getItem(`${this.realm}.sessionUuid`)
        if(uuid === null) {
            uuid = generateSessionUuid()
            localStorage.setItem(`${this.realm}.sessionUuid`, uuid)
        }

        return uuid
    }

    /**
     * Return if this is the first time the application was started
     */
    get firstStart() {
        const firstStart = localStorage.getItem(`${this.realm}.firstStart`)
        // firstStart === null: If the key firstStart doesn't even exist, we certainly started for the first time
        // firstStart === "true": Local storage can only store strings, so this is our way of casting to boolean
        return firstStart === null || firstStart === "true"
    }

    /**
     * Set if this is the first time the application was started
     */
    set firstStart(value) {
        if (typeof value !== "boolean") {
            throw new Error("value must be a boolean")
        }

        localStorage.setItem(`${this.realm}.firstStart`, value)
    }

    /**
     * Get the index of the exercise the user is currently working on
     */
    get exerciseIndex() {
        const exerciseIndex = localStorage.getItem(`${this.realm}.exerciseIndex`)
        if (exerciseIndex === null) {
            return null
        } else {
            return parseInt(exerciseIndex, 10)
        }
    }

    /**
     * Set the index of the exercise the user is currently working on.
     * 0 <= index < exercises.length
     */
    set exerciseIndex(index) {
        if (typeof index !== "number") {
            throw new Error("index must be a number")
        }

        // TODO Verify that index is in range, else
        // throw new Error(`Exercise index must be >= 0 and < exercises.length (${pooltest.exercises.length})`)

        localStorage.setItem(`${this.realm}.exerciseIndex`, index)
    }

    /**
     * Get a list of indices of all solved exercises
     */
    get solvedExercises() {
        return JSON.parse(localStorage.getItem(`${this.realm}.solvedExercises`)) || []
    }

    /**
     * Private setter for the list of all solved exercises
     * @param {Array} indices Array of indicies of exercises that are solved
     */
    _setSolvedExercises(indices) {
        localStorage.setItem(`${this.realm}.solvedExercises`, JSON.stringify(indices))
    }

    /**
     * Mark the given exercise as solved
     * @param {number} exerciseIndex The index of the exercise to be marked as solved
     */
    markAsSolved(exerciseIndex) {
        if (this.solvedExercises.includes(exerciseIndex)) {
            throw new Error(`Can't solve the same exercise (${exerciseIndex}) twice`)
        }

        // TODO Verify that exerciseIndex is in range, else
        // throw new Error(`Exercise index must be >= 0 and < exercises.length (${pooltest.exercises.length})`)

        const newIndices = this.solvedExercises
        newIndices.push(exerciseIndex)
        this._setSolvedExercises(newIndices)
    }

    /**
     * Return a string representation of the current state
     */
    toString() {
        return JSON.stringify({
            "realm": this.realm,
            "sessionUuid": this.sessionUuid,
            "firstStart": this.firstStart,
            "exerciseIndex": this.exerciseIndex,
            "solvedExercises": this.solvedExercises
        })
    }
}
