import { State } from "./state.mjs"
import * as tests from "./exercises.mjs"
import { createExerciseCard, displayAsSolved, displayAsNonCurrent, displayAsCurrent, displayAsVerifying, displayAsNonVerifying, displayAsUnusuallyLongVerifying } from "./exercises.cards.mjs"
import { Exercise } from "./exercises.api.mjs"
import { isVmLoading, onCopyOrPasteHotkey, setAutoFocusEnabled } from "./jslinux.api.mjs"
import { findParent, waitFor } from "./util.mjs"
import { initDontShowAgain, setDialogShown } from "./dialogs.mjs"
import { setTooltipShown } from "./tooltips.mjs"

const testId = new URLSearchParams(location.search).get("id")
const currentTest = tests[testId]
const currentState = new State(testId)

/**
 * Create a DOM card for the given exercise, displays is as solved if solved and inits event listeners
 * @param {Exercise} exercise The exercise
 * @returns The DOM card
 */
function createAndInitExerciseCard(exercise) {
    const cardEl = createExerciseCard(exercise, () => {
        if (!exercise.isSolved(currentState) && !exercise.isCurrent(currentState)) {
            startExercise(exercise)
        }
    })

    if (exercise.isSolved(currentState)) {
        displayAsSolved(cardEl)
    }

    return cardEl
}

/**
 * Load all exercises into the #exercises list in the DOM (in a collapsed state, 
 * meaning that description and action buttons are not visible).
 */
function initExerciseList() {
    document.querySelector("#exercises").append(...currentTest.exercises.map(createAndInitExerciseCard))
}

/**
 * Start the next exercise or show the hand in tooltip, if we're done
 */
function startNextExerciseOrShowCompletionTooltip() {
    const nextExercise = currentTest.getNextUnsolvedExercise(currentState)
    startExercise(nextExercise)
    if (nextExercise === null) {
        setTimeout(()=>{
            setTooltipShown(currentTest.handInEnabled ? "hand-in-tooltip" : "go-home-tooltip", true)
        }, 500)
    }
}

/**
 * Start an exercise, also updating the currentState.
 *
 * If no exercise is given, none is started. In this case the function only has 
 * the effect of marking all exercises as not-current (i.e. collapsing them).
 * @param {Exercise} exercise Exercise to start
 */
function startExercise(exercise) {
    // Display all exercises as non-current, except for this one
    document.querySelectorAll(".exercise").forEach(el => displayAsNonCurrent(el))

    if(!exercise) return
    
    const cardEl = document.querySelector(`.exercise[data-exercise-index='${exercise.index}']`)
    let unusuallyLongTimeoutId = null

    exercise.markAsCurrent(currentState)
    exercise.onDescribed((/** @type {string} */ description) => cardEl.querySelector(".description").textContent = description)
    exercise.onVerifying(()=>{
        displayAsVerifying(cardEl)
        unusuallyLongTimeoutId = setTimeout(()=>{
            displayAsUnusuallyLongVerifying(cardEl)
        }, 5000)
    })
    exercise.onVerified(
        (/** @type {boolean} */ success) => {
            displayAsNonVerifying(cardEl)
            if(unusuallyLongTimeoutId) {
                clearTimeout(unusuallyLongTimeoutId)
            }

            if (success) {
                exercise.markAsSolved(currentState)
                displayAsSolved(cardEl)
                startNextExerciseOrShowCompletionTooltip()
            } else {
                exercise.retry()
                setDialogShown("failed-attempt-dialog", true)
            }
        }
    )

    exercise.start()
    displayAsCurrent(cardEl)
}

function initActionLinks() {
    // TODO Replace with something good-looking
    document.querySelectorAll("[data-action]").forEach(el => {
        const handler = {
            "hand-in": function () {
                setAutoFocusEnabled(false)
                const handInToken = currentTest.getHandInToken(currentState)
                navigator.clipboard.writeText(handInToken)
                document.querySelector("#hand-in-token").textContent = handInToken
                setTooltipShown("hand-in-tooltip", false)
                setDialogShown("hand-in-dialog", true)
            },
            "copy-token": function () {
                const handInToken = currentTest.getHandInToken(currentState)
                navigator.clipboard.writeText(handInToken)
                document.querySelector("#copy-token .bi").classList.replace("bi-clipboard", "bi-clipboard-check")
                setTimeout(()=>{
                    document.querySelector("#copy-token .bi").classList.replace("bi-clipboard-check", "bi-clipboard")
                }, 700)
            },
            "begin": function () {
                setDialogShown("welcome-dialog", false)
                currentTest.markWelcomeAsRead(currentState)
            },
            "close": function () {
                setDialogShown(findParent(el, ".dialog").id, false)
            },
            "complete": function () {
                setDialogShown("hand-in-dialog", false)
                setAutoFocusEnabled(true)
            },
            "welcome-again": function () {
                document.querySelector("#welcome-dialog [data-action='begin']").style.display = "none"
                document.querySelector("#welcome-dialog [data-action='close']").style.display = "block"
                setDialogShown("welcome-dialog", true)
            },
            "reset": function () {
                const response = prompt("Möchtest du wirklich allen Fortschritt löschen und von vorne beginnen? Achtung: Die Seite verhält sich danach, als würdest du sie zum allerersten Mal besuchen (insbesondere: falls du schon angefangen hast, den Pooltest zu bearbeiten, ist dein Fortschritt komplett weg). Gib 'ZURÜCKSETZEN' (ohne Anführungszeichen) ein, falls du alles zurücksetzen willst. Ansonsten drücke einfach auf Abbrechen :)")
                if(response === "ZURÜCKSETZEN") {
                    localStorage.clear()
                    location.reload()
                }
            }
        }[el.dataset.action]

        el.onclick = handler
    })
}

/**
 * Enables or disables all exercise action buttons
 * @param {boolean} enabled If all exercise action buttons should be enabled
 */
function setExerciseActionButtonsEnabled(enabled) {
    document.querySelectorAll(".exercise.card .actions button").forEach(buttonEl => buttonEl.disabled = !enabled)
}

/**
 * Event handler for unhandled errors
 * @param {Event} event The error event
 */
function onUnhandledError(event) {
    // TODO Use something nicer
    alert("Es ist ein Programmierfehler in YALIKEJAZZ aufgetreten, Details stehen in der Entwicklerkonsole. Bitte lade die Seite neu; dein Fortschritt bleibt erhalten. Außerdem wären wir dir sehr dankbar für eine Fehlermeldung an technik@cl.uni-heidelberg.de.")
}

function redirectToPractice() {
    // TODO Use something nicer based on URLSearchParams
    location.search = `?id=practice`
}

function initWelcomeDialog() {
    if(currentTest.welcomeHtml) {
        document.querySelector("#welcome").innerHTML = currentTest.welcomeHtml

        const exerciseCnt = document.querySelector("#exerciseCnt")
        if (exerciseCnt) {
            exerciseCnt.textContent = currentTest.exercises.length
        }

        if (!currentTest.wasWelcomeRead(currentState)) {
            setDialogShown("welcome-dialog", true)
        }
    } else {
        // No welcome dialog exists, so we also don't want the user to be able to show it again
        // (because what does this even mean)
        document.querySelector("[data-action='welcome-again']").style.display = "none"
    }
}

export async function main() {
    if(currentTest === undefined) {
        // When no test can be loaded, redirect to a URL which leads to a known existing test
        redirectToPractice()
    }

    document.querySelector("#testName").textContent = currentTest.name
    document.title = `${currentTest.name} · ${document.title}`

    if(!currentTest.handInEnabled) {
        document.querySelector("[data-action='hand-in']").style.display = "none"
    }

    initWelcomeDialog()
    initDontShowAgain()
    initExerciseList()
    initActionLinks()

    setAutoFocusEnabled(true)
    setExerciseActionButtonsEnabled(false)
    onCopyOrPasteHotkey(()=>setDialogShown("copy-paste-dialog", true))
    
    startNextExerciseOrShowCompletionTooltip()

    // Enable all exercise action buttons as soon as the VM has booted
    await waitFor(()=>!isVmLoading())
    setExerciseActionButtonsEnabled(true)
}

window.onload = main
window.addEventListener("unhandledrejection", onUnhandledError)
window.addEventListener("event", onUnhandledError)
